"""Module that controls the overall state of the program screen"""

import datetime
import jdatetime

from calcure.data import Events, AppState, CalState
from calcure.calendars import Calendar


class Screen:
    """Main state of the program that describes what is displayed and how"""
    def __init__(self, stdscr, cf):
        self.stdscr = stdscr
        self.privacy = cf.PRIVACY_MODE
        self.state = cf.DEFAULT_VIEW
        self.calendar_state = CalState.DAILY if cf.DEFAULT_CALENDAR_VIEW == "daily" else CalState.MONTHLY
        self.use_persian_calendar = cf.USE_PERSIAN_CALENDAR
        self.split = cf.SPLIT_SCREEN
        self.right_pane_percentage = cf.RIGHT_PANE_PERCENTAGE
        self.currently_drawn = self.state
        self.selection_mode = False
        self.refresh_now = True
        self.key = None
        self.day = self.today.day
        self.month = self.today.month
        self.year = self.today.year

    @property
    def is_active_pane(self):
        """Return True if currently drawn pane in the active one"""
        return self.state == self.currently_drawn

    @property
    def y_max(self):
        """Get maximum size of the screen"""
        y_max, _ = self.stdscr.getmaxyx()
        return y_max

    @property
    def journal_pane_width(self):
        """Calculate the width of the right pane if the value is adequate"""
        _, x_max = self.stdscr.getmaxyx()
        if 5 < self.right_pane_percentage < 95:
            return int(x_max//(100/self.right_pane_percentage))
        return x_max//4

    @property
    def x_max(self):
        """Calculate the right boundary of the screen"""
        _, x_max = self.stdscr.getmaxyx()
        if x_max < 40:
            self.split = False
        if self.split and self.currently_drawn != AppState.JOURNAL:
            return x_max - self.journal_pane_width
        return x_max

    @property
    def x_min(self):
        """Calculate the left boundary of the screen"""
        _, x_max = self.stdscr.getmaxyx()
        if x_max < self.journal_pane_width:
            self.split = False
        if self.split and self.currently_drawn == AppState.JOURNAL:
            return x_max - self.journal_pane_width + 2
        return 0

    @property
    def date(self) -> datetime:
        """Return displayed date in datetime format"""
        if self.use_persian_calendar:
            return jdatetime.date(self.year, self.month, self.day)
        else:
            return datetime.date(self.year, self.month, self.day)

    @property
    def today(self) -> datetime:
        """Return todays's date in datetime format"""
        if self.use_persian_calendar:
            return jdatetime.date.today()
        else:
            return datetime.date.today()

    def next_month(self):
        """Switches to the next month"""
        if self.month < 12:
            self.month += 1
        else:
            self.month = 1
            self.year += 1

    def previous_month(self):
        """Switches to the previous month"""
        if self.month > 1:
            self.month -= 1
        else:
            self.month = 12
            self.year -= 1

    def next_day(self):
        """Switch to the next day"""
        days_in_this_month = Calendar(0, self.use_persian_calendar).last_day(self.year, self.month)
        if self.day < days_in_this_month:
            self.day += 1
        else:
            self.day = 1
            if self.month < 12:
                self.month += 1
            else:
                self.month = 1
                self.year += 1

    def previous_day(self):
        """Switch to the previous day"""
        if self.day > 1:
            self.day -= 1
        else:
            if self.month > 1:
                self.month -= 1
            else:
                self.month = 12
                self.year -= 1
            self.day = Calendar(0, self.use_persian_calendar).last_day(self.year, self.month)

    def reset_to_today(self):
        """Reset the day, month, and year to the current date"""
        self.month = self.today.month
        self.year = self.today.year
        self.day = self.today.day

    def is_valid_day(self, number) -> bool:
        """Check if input corresponds to a date in this month"""
        if number is None:
            return False
        return 0 < number <= Calendar(0, self.use_persian_calendar).last_day(self.year, self.month)


    def is_valid_date(self, year, month, day) -> bool:
        """Check if a date corresponds to any actually existing date"""
        if None in [year, month, day]:
            return False
        if 1 <= month <= 12:
            return 0 < day <= Calendar(0, self.use_persian_calendar).last_day(year, month)
        return False



"""Module that controls loading data from files and libraries"""

import configparser
import pathlib
import csv
import os
import datetime
import ics
import urllib.request
import io
import logging

from calcure.data import *
from calcure.calendars import convert_to_persian_date, convert_to_gregorian_date


class LoaderCSV:
    """Load data from CSV files"""

    def create_file(self, filename):
        """Create CSV file"""
        try:
            with open(filename, "w+", encoding="utf-8") as file:
                pass
            return []
        except (FileNotFoundError, NameError) as e_message:
            logging.error("Problem occured trying to create %s. %s", filename, e_message)
            return []

    def read_file(self, filename):
        """Read CSV file or create new one if it does not exist"""
        try:
            with open(filename, "r", encoding="utf-8") as file:
                lines = csv.reader(file, delimiter = ',')
                return list(lines)
        except IOError: # File does not exist
            logging.info("Creating %s.", filename)
            return self.create_file(filename)


class TaskLoaderCSV(LoaderCSV):
    """Load tasks from CSV files"""

    def __init__(self, cf):
        self.user_tasks = Tasks()
        self.tasks_file = cf.TASKS_FILE
        self.use_persian_calendar = cf.USE_PERSIAN_CALENDAR

    @property
    def is_task_format_old(self):
        """Check if the database format is old"""
        with open(self.tasks_file, "r", encoding="utf-8") as f:
            text = f.read()
        return text[0] == '"'

    def load(self):
        """Reads from CSV file"""
        lines = self.read_file(self.tasks_file)

        for index, row in enumerate(lines):
            task_id = index

            # Read task dates:
            if self.is_task_format_old:
                shift = 0
                year = 0
                month = 0
                day = 0
            else:
                shift = 3
                year = int(row[0])
                month = int(row[1])
                day = int(row[2])

            # Convert to persian date if needed and if it is not zero date:
            if self.use_persian_calendar and year != 0:
                year, month, day = convert_to_persian_date(year, month, day)

            # Read task name and statuses:
            if row[0 + shift][0] == '.':
                name = row[0 + shift][1:]
                is_private = True
            else:
                name = row[0 + shift]
                is_private = False
            status = Status[row[1 + shift].upper()]
            stamps = row[(2 + shift):] if len(row) > 2 else []
            timer = Timer(stamps)

            # Add task:
            new_task = Task(task_id, name, status, timer, is_private, year, month, day)
            self.user_tasks.add_item(new_task)
        return self.user_tasks


class EventLoaderCSV(LoaderCSV):
    """Load events from CSV files"""

    def __init__(self, cf):
        self.user_events = Events()
        self.events_file = cf.EVENTS_FILE
        self.use_persian_calendar = cf.USE_PERSIAN_CALENDAR

    def load(self):
        """Read from CSV file"""
        lines = self.read_file(self.events_file)
        for index, row in enumerate(lines):
            event_id = index
            year = int(row[1])
            month = int(row[2])
            day = int(row[3])
            if row[4][0] == '.':
                name = row[4][1:]
                is_private = True
            else:
                name = row[4]
                is_private = False

            # Account for old versions of the datafile:
            if len(row) > 5:
                repetition = int(row[5])
                if row[6] == 'd':
                    frequency = Frequency.DAILY
                elif row[6] == 'w':
                    frequency = Frequency.WEEKLY
                elif row[6] == 'm':
                    frequency = Frequency.MONTHLY
                elif row[6] == 'y':
                    frequency = Frequency.YEARLY
                else:
                    try:
                        frequency = Frequency[row[6].upper()]
                    except (ValueError, KeyError):
                        frequency = Frequency.ONCE
            else:
                repetition = '1'
                frequency = Frequency.ONCE
            if len(row) > 7:
                status = Status[row[7].upper()]
            else:
                status = Status.NORMAL

            # Convert to persian date if needed:
            if self.use_persian_calendar:
                year, month, day = convert_to_persian_date(year, month, day)

            # Add event:
            new_event = UserEvent(event_id, year, month, day, name, repetition, frequency, status, is_private)
            self.user_events.add_item(new_event)
        return self.user_events


class HolidayLoader:
    """Load holidays for this country around this year"""

    def __init__(self, cf):
        self.holidays = Events()
        self.country = cf.HOLIDAY_COUNTRY
        self.use_persian_calendar = cf.USE_PERSIAN_CALENDAR

    def load(self):
        """Load list of holidays from 'holidays' module"""
        try:
            import holidays as hl
            year = datetime.date.today().year
            holiday_events = eval("hl."+self.country+"(years=[year-2, year-1, year, year+1, year+2, year+3, year+4])")
            for date, name in holiday_events.items():

                # Convert to persian date if needed:
                if self.use_persian_calendar:
                    year, month, day = convert_to_persian_date(date.year, date.month, date.day)
                else:
                    year, month, day = date.year, date.month, date.day

                # Add holiday:
                holiday = Event(year, month, day, name)
                self.holidays.add_item(holiday)

        except ModuleNotFoundError:
            logging.error("Couldn't load holidays. Module holydays is not installed. Try 'pip install holydays'")
            pass
        except (SyntaxError, AttributeError) as e_message:
            logging.error("Couldn't load holidays. Country might be incorrect. %s", e_message)
            pass
        return self.holidays


class BirthdayLoader:
    """Load birthdays of contacts"""

    def __init__(self, cf):
        self.birthdays = Birthdays()
        self.abook_file = str(pathlib.Path.home())+"/.abook/addressbook"
        self.use_persian_calendar = cf.USE_PERSIAN_CALENDAR

    def load(self):
        """Loading birthdays from abook contacts"""

        # Quit if file does not exists:
        if not os.path.exists(self.abook_file):
            logging.warning("Couldn't load birthdays. File. %s does not exist.", self.abook_file)
            return self.birthdays

        abook = configparser.ConfigParser()
        abook.read(self.abook_file)
        for each_contact in abook.sections():
            for key, _ in abook.items(each_contact):
                if key in ["birthday", "anniversary"]:
                    month = int(abook[each_contact][key][-5:-3])
                    day = int(abook[each_contact][key][-2:])
                    name = abook[each_contact]["name"]

                    # Convert to persian date if needed:
                    if self.use_persian_calendar:
                        _, month, day = convert_to_persian_date(1000, month, day)

                    # Add birthday:
                    birthday = Event(1, month, day, name)
                    self.birthdays.add_item(birthday)
        return self.birthdays


class LoaderICS:
    """Load data from ICS files"""

    def read_lines(self, file):
        """Read the file line-by-line and remove multiple PRODID lines"""
        already_has_prodid = False
        text = ""
        for line in file:
            # If there is more than one PRODID line or a TZUNTIL line, skip them:
            if ((not (already_has_prodid and "PRODID:" in line)) and
                (not "TZUNTIL" in line)):
                text += line

            if "PRODID:" in line:
                already_has_prodid = True
        return text

    def read_file(self, path):
        """Parse an ics file if it exists"""
        if not os.path.exists(path):
            logging.error("Failed to load %s because file does not exist.", path)
            return ""
        with open(path, 'r', encoding="utf-8") as file:
            return self.read_lines(file)

    def read_url(self, path):
        """Parse an ics URL if it exists and networks works"""
        try:
            with urllib.request.urlopen(path) as response:
                file = io.TextIOWrapper(response, 'utf-8')
                return self.read_lines(file)
        except urllib.error.HTTPError as e_message:
            logging.error("Failed to load from %s. Probably url is wrong. %s", path, e_message)
            return ""
        except urllib.error.URLError as e_message:
            logging.error("Failed to load from %s. Probably no internet connection. %s", path, e_message)
            return ""

    def read_resource(self, path):
        """Determine type of the resourse, parse it, and return list of strings for each file"""
        ics_files = []

        # If it's a URL, try to load it:
        if path.startswith('http'):
            ics_files.append(self.read_url(path))
            return ics_files

        # If it's a local file, read it:
        if path.endswith('.ics'):
            ics_files.append(self.read_file(path))
            return ics_files

        # Otherwise, assume it's a folder, and read every file inside:
        for root, directories, files in os.walk(path):
            for filename in files:
                # Get the full path to the file
                file_path = os.path.join(root, filename)

                # `path` may contain files with metadata, e.g. `color` and
                # `displayname`. For now, exclude those while loading.
                if file_path.endswith('.ics'):
                    ics_files.append(self.read_file(file_path))

        return ics_files


class TaskLoaderICS(LoaderICS):
    """Load tasks from ICS files"""

    def __init__(self, cf):
        self.user_ics_tasks = Tasks()
        self.ics_task_files = cf.ICS_TASK_FILES
        self.use_persian_calendar = cf.USE_PERSIAN_CALENDAR

    def parse_task(self, task, calendar_number):
        """Parse single task and add it to the user_ics_tasks"""

        if task.status == "CANCELLED":
            return

        task_id = self.user_ics_tasks.generate_id()

        # Assign status from priority:
        status = Status.NORMAL
        if task.priority is not None:
            if task.priority > 5:
                status = Status.UNIMPORTANT
            if task.priority < 5:
                status = Status.IMPORTANT

        # Correct according to status:
        if task.status == "COMPLETED":
            status = Status.DONE

        name = task.name

        # Try reading task due date:
        try:
            year = task.due.year
            month = task.due.month
            day = task.due.day
        except AttributeError:
            year, month, day = 0, 0, 0

        timer = Timer([])
        is_private = False

        # Add task:
        new_task = Task(task_id, name, status, timer, is_private,
                        year, month, day, calendar_number)
        self.user_ics_tasks.add_item(new_task)


    def load(self):
        """Load tasks from each of the ics files"""

        # Quit if the files are not specified in config:
        if self.ics_task_files is None:
            return self.user_ics_tasks

        for calendar_number, filename in enumerate(self.ics_task_files):
            # For each resourse from config, load a list that has one or more ics files:
            ics_files = self.read_resource(filename)
            for ics_file in ics_files:
                try:
                    cal = ics.Calendar(ics_file)
                    for task in cal.todos:
                        self.parse_task(task, calendar_number)
                except Exception as e_message:
                    logging.error("Failed to parse %s. %s", filename, e_message)

        return self.user_ics_tasks


class EventLoaderICS(LoaderICS):
    """Load events from ICS files"""

    def __init__(self, cf):
        self.user_ics_events = Events()
        self.ics_event_files = cf.ICS_EVENT_FILES
        self.use_persian_calendar = cf.USE_PERSIAN_CALENDAR

    def parse_event(self, event, index, calendar_number):
        """Parse singe event and add it to user_ics_events"""

        # Default parameters:
        event_id = index
        repetition = '1'
        frequency = Frequency.ONCE
        status = Status.NORMAL
        is_private = False

        # Parameters of the event from ics file, if they exist:
        name = event.name if event.name is not None else ""
        all_day = event.all_day if event.all_day is not None else True
        year = event.begin.year if event.begin else 0
        month = event.begin.month if event.begin else 1
        day = event.begin.day if event.begin else 1

        # Add start time to the name of non-all-day events:
        if not all_day:
            hour = event.begin.hour if event.begin else 0
            minute = event.begin.minute if event.begin else 0
            name = f"{hour:0=2}:{minute:0=2} {name}"

        # Convert to persian date if needed:
        if self.use_persian_calendar:
            year, month, day = convert_to_persian_date(year, month, day)

        # Add event:
        new_event = UserEvent(event_id, year, month, day, name, repetition,
                              frequency, status, is_private, calendar_number)
        self.user_ics_events.add_item(new_event)

    def load(self):
        """Load events from each of the ics files"""

        # Quit if the files are not specified in config:
        if self.ics_event_files is None:
            return self.user_ics_events

        for calendar_number, filename in enumerate(self.ics_event_files):
            # For each resourse from config, load a list that has one or more ics files:
            ics_files = self.read_resource(filename)
            for ics_file in ics_files:
                try:
                    cal = ics.Calendar(ics_file)
                    for index, event in enumerate(cal.events):
                        self.parse_event(event, index, calendar_number)
                except Exception as e_message:
                    logging.error("Failed to parse %s. %s", filename, e_message)
        return self.user_ics_events



"""Module that provides few converter functions"""

import curses
import jdatetime
from enum import Enum, auto


class Color(Enum):
    """Colors read from user config"""
    DAY_NAMES = auto()
    WEEKENDS = auto()
    HINTS = auto()
    TODAY = auto()
    DAYS = auto()
    WEEKEND_NAMES = auto()
    BIRTHDAYS = auto()
    PROMPTS = auto()
    CONFIRMATIONS = auto()
    TITLE = auto()
    TODO = auto()
    DONE = auto()
    IMPORTANT = auto()
    TIMER = auto()
    TIMER_PAUSED = auto()
    HOLIDAYS = auto()
    EVENTS = auto()
    TIME = auto()
    WEATHER = auto()
    UNIMPORTANT = auto()
    CALENDAR_HEADER = auto()
    ACTIVE_PANE = auto()
    SEPARATOR = auto()
    EMPTY = auto()
    CALENDAR_BOARDER = auto()
    DEADLINES = auto()
    ICS_CALENDARS0 = auto()
    ICS_CALENDARS1 = auto()
    ICS_CALENDARS2 = auto()
    ICS_CALENDARS3 = auto()
    ICS_CALENDARS4 = auto()
    ICS_CALENDARS5 = auto()
    ICS_CALENDARS6 = auto()
    ICS_CALENDARS7 = auto()
    ICS_CALENDARS8 = auto()
    ICS_CALENDARS9 = auto()


def initialize_colors(cf):
    """Define all the color pairs"""
    curses.start_color()
    curses.use_default_colors()
    curses.init_pair(Color.DAY_NAMES.value, cf.COLOR_DAY_NAMES, cf.COLOR_BACKGROUND)
    curses.init_pair(Color.WEEKENDS.value, cf.COLOR_WEEKENDS, cf.COLOR_BACKGROUND)
    curses.init_pair(Color.HINTS.value, cf.COLOR_HINTS, cf.COLOR_BACKGROUND)
    curses.init_pair(Color.TODAY.value, cf.COLOR_TODAY, cf.COLOR_BACKGROUND)
    curses.init_pair(Color.DAYS.value, cf.COLOR_DAYS, cf.COLOR_BACKGROUND)
    curses.init_pair(Color.WEEKEND_NAMES.value, cf.COLOR_WEEKEND_NAMES, cf.COLOR_BACKGROUND)
    curses.init_pair(Color.BIRTHDAYS.value, cf.COLOR_BIRTHDAYS, cf.COLOR_BACKGROUND)
    curses.init_pair(Color.PROMPTS.value, cf.COLOR_PROMPTS, cf.COLOR_BACKGROUND)
    curses.init_pair(Color.CONFIRMATIONS.value, cf.COLOR_CONFIRMATIONS, cf.COLOR_BACKGROUND)
    curses.init_pair(Color.TITLE.value, cf.COLOR_TITLE, cf.COLOR_BACKGROUND)
    curses.init_pair(Color.TODO.value, cf.COLOR_TODO, cf.COLOR_BACKGROUND)
    curses.init_pair(Color.DONE.value, cf.COLOR_DONE, cf.COLOR_BACKGROUND)
    curses.init_pair(Color.IMPORTANT.value, cf.COLOR_IMPORTANT, cf.COLOR_BACKGROUND)
    curses.init_pair(Color.TIMER.value, cf.COLOR_TIMER, cf.COLOR_BACKGROUND)
    curses.init_pair(Color.TIMER_PAUSED.value, cf.COLOR_TIMER_PAUSED, cf.COLOR_BACKGROUND)
    curses.init_pair(Color.HOLIDAYS.value, cf.COLOR_HOLIDAYS, cf.COLOR_BACKGROUND)
    curses.init_pair(Color.EVENTS.value, cf.COLOR_EVENTS, cf.COLOR_BACKGROUND)
    curses.init_pair(Color.TIME.value, cf.COLOR_TIME, cf.COLOR_BACKGROUND)
    curses.init_pair(Color.WEATHER.value, cf.COLOR_WEATHER, cf.COLOR_BACKGROUND)
    curses.init_pair(Color.UNIMPORTANT.value, cf.COLOR_UNIMPORTANT, cf.COLOR_BACKGROUND)
    curses.init_pair(Color.CALENDAR_HEADER.value, cf.COLOR_CALENDAR_HEADER, cf.COLOR_BACKGROUND)
    curses.init_pair(Color.ACTIVE_PANE.value, cf.COLOR_ACTIVE_PANE, cf.COLOR_BACKGROUND)
    curses.init_pair(Color.SEPARATOR.value, cf.COLOR_SEPARATOR, cf.COLOR_BACKGROUND)
    curses.init_pair(Color.EMPTY.value, cf.COLOR_BACKGROUND, cf.COLOR_BACKGROUND)
    curses.init_pair(Color.CALENDAR_BOARDER.value, cf.COLOR_CALENDAR_BOARDER, cf.COLOR_BACKGROUND)
    curses.init_pair(Color.DEADLINES.value, cf.COLOR_DEADLINES, cf.COLOR_BACKGROUND)

    if not cf.MINIMAL_WEEKEND_INDICATOR:
        curses.init_pair(Color.WEEKENDS.value, curses.COLOR_BLACK, cf.COLOR_WEEKENDS)
    if not cf.MINIMAL_TODAY_INDICATOR:
        curses.init_pair(Color.TODAY.value, curses.COLOR_BLACK, cf.COLOR_TODAY)
    if not cf.MINIMAL_DAYS_INDICATOR:
        curses.init_pair(Color.DAYS.value, curses.COLOR_BLACK, cf.COLOR_DAYS)

    # Assign color pair for each ics resourse:
    if cf.ICS_EVENT_FILES is None:
        return

    for index in range(len(cf.ICS_EVENT_FILES)):
        if index < len(cf.COLOR_ICS_CALENDARS):
            color = cf.COLOR_ICS_CALENDARS[index] # Take colors from config
        else:
            color = cf.COLOR_ICS_CALENDARS[-1] # Remaining resourses assume the last assigned color
        curses.init_pair(Color.ICS_CALENDARS0.value + index, color, cf.COLOR_BACKGROUND)



"""Module that controls the weather storing and loading"""

import subprocess
import logging


class Weather:
    """Information about the weather today"""

    def __init__(self, city, is_metric_units):
        self.city = city
        self.forcast = ""
        self.max_load_time = 2  # seconds
        self.units = "?m" if is_metric_units else "?u"

    def load_from_wttr(self):
        """Load the weather info from wttr.in"""
        try:
            request_url = f"wttr.in/{self.city}{self.units}&format=3"
            self.forcast = str(subprocess.check_output(["curl", "-s", request_url],
                                                        timeout=self.max_load_time,
                                                        encoding='utf-8'))[:-1]
            self.forcast = self.forcast.split(':')[1]
        except (subprocess.TimeoutExpired, subprocess.CalledProcessError, IndexError):
            logging.warning("Weather failed to load.")
            self.forcast = ""



"""Module that provides methods to import data from other programs"""
import logging

from calcure.data import *
from calcure.calendars import convert_to_persian_date


class Importer:
    """Import tasks and events from files of other programs"""

    def __init__(self, user_tasks, user_events, cf):
        self.user_tasks = user_tasks
        self.user_events = user_events
        self.tasks_file = cf.TASKS_FILE
        self.events_file = cf.EVENTS_FILE
        self.calcurse_todo_file = cf.CALCURSE_TODO_FILE
        self.calcurse_events_file = cf.CALCURSE_EVENTS_FILE
        self.taskwarrior_folder = cf.TASKWARRIOR_FOLDER
        self.use_persian_calendar = cf.USE_PERSIAN_CALENDAR

    def read_file(self, filename):
        """Try to read a file and return its lines"""
        try:
            with open(filename, "r", encoding="utf-8") as file:
                lines = file.readlines()
            return lines
        except (IOError, FileNotFoundError, NameError):
            logging.error("Problem occurred accessing %s.", filename)
            return []

    def import_tasks_from_calcurse(self):
        """Import tasks from calcurse database"""
        lines = self.read_file(self.calcurse_todo_file)
        for line in lines:
            name = line[4:-1]
            importance = line[1]
            if (len(name) > 0) and not self.user_tasks.item_exists(name):
                if importance in ['1', '2']:
                    status = Status.IMPORTANT
                elif importance in ['8', '9', '10']:
                    status = Status.UNIMPORTANT
                else:
                    status = Status.NORMAL
                task_id = self.user_tasks.generate_id()
                is_private = False
                timer = Timer([])
                self.user_tasks.add_item(Task(task_id, name, status, timer, is_private))

    def import_tasks_from_taskwarrior(self):
        """Import tasks from taskwarrior database"""
        lines = self.read_file(self.taskwarrior_folder+"/pending.data")
        for line in lines:
            if len(line) > 0:
                name = line.split('description:"', 1)[1]
                name = name.split('"', 1)[0]
                if not self.user_tasks.item_exists(name):
                    task_id = self.user_tasks.generate_id()
                    is_private = False
                    timer = Timer([])
                    status = Status.NORMAL
                    self.user_tasks.add_item(Task(task_id, name, status, timer, is_private))

    def import_events_from_calcurse(self):
        """Importing events from calcurse apt file into our events file"""
        lines = self.read_file(self.calcurse_events_file)
        for line in lines:
            month = int(line[0:2])
            day = int(line[3:5])
            year = int(line[6:10])
            if line[11] == "[":
                name = line[15:-1]
            elif line[11] == "@":
                name = line[35:-1]
                name = name.replace('|',' ')
            else:
                name = ''
            if not self.user_events.items:
                event_id = 0
            else:
                event_id = self.user_events.items[-1].item_id + 1
            is_private = False

            # Convert to persian date if needed:
            if self.use_persian_calendar:
                year, month, day = convert_to_persian_date(year, month, day)

            imported_event = UserEvent(event_id, year, month, day, name, 1,
                                       Frequency.ONCE, Status.NORMAL, is_private)
            if not self.user_events.event_exists(imported_event):
                self.user_events.add_item(imported_event)



"""Module that controls logging and error output"""

import logging
import io

from calcure.configuration import cf


class Error():
    """Error messages displayed to the user.
    Different errors have different purposes:
    ERROR type is reserved for loading errors;
    WARNING type is reserved for incorrect user input;"""

    def __init__(self):
        self.buffer = io.StringIO()
        self.file = f"{cf.config_folder}/info.log"

    @property
    def has_occured(self):
        """Has any errors occured?"""
        return self.buffer.getvalue() != ""

    @property
    def number_of_errors(self):
        """Return the number of different errors in the buffer"""
        return self.buffer.getvalue().count("[")

    @property
    def text(self):
        """Return the string with the text of the error"""
        return self.buffer.getvalue().split("] ")[1]

    @property
    def type(self):
        """Return the string with type of the error"""
        return self.buffer.getvalue().split("] ")[0]

    def clear_buffer(self):
        """Clear the buffer containing errors"""
        self.buffer.seek(0)
        self.buffer.truncate(0)


# Initialise error:
error = Error()

# Start logging:
logging.basicConfig(level=logging.INFO,
                    format="[%(levelname)s] %(message)s",
                    # encoding='utf-8',
                    handlers=[logging.FileHandler(error.file, 'w'),
                              logging.StreamHandler(),
                              logging.StreamHandler(error.buffer),
                              ])



"""Turkish translations of the program interface"""

MSG_WELCOME_1 = "Calcure'a hoş geldiniz"
MSG_WELCOME_2 = "Terminal takviminiz ve görev yöneticiniz!"
MSG_WELCOME_3 = "Yapılandırma ve veri dosyaları şu adreste oluşturuldu:"
MSG_WELCOME_4 = "Destek, katkı ve ek bilgi için şu adresi ziyaret edin:"
MSG_WELCOME_5 = "Tüm tuş atamalarını görmek için ? tuşuna veya devam etmek için başka bir tuşa basın."

TITLE_KEYS_GENERAL  = "GENEL TUŞ ATAMALARI"
TITLE_KEYS_CALENDAR = "TAKVIM TUŞ ATAMALARI"
TITLE_KEYS_JOURNAL  = "GÜNLÜK TUŞ ATAMALARI"

KEYS_GENERAL = {
        " Space ": "Takvim ve günlük arasında geçiş yapma",
        "   /   ": "Bölünmüş ekranı aç / kapat",
        "   *   ": "Genel gizliliği aç / kapat",
        "   ?   ": "Bu yardımı aç / kapat",
        "   q   ": "Çıkış",
        }

KEYS_CALENDAR = {
        "  a(A) ": "Etkinlik (yinelenen) ekleme",
        "  n,🠒  ": "Gelecek ay (gün)",
        "  p,🠐  ": "Önceki ay (gün)",
        "   x   ": "Etkinlik silme",
        "  e,r  ": "Etkinlik düzenleme",
        "  m(M) ": "Etkinlik taşıma",
        "  g(G) ": "Belirli bir güne git",
        "   v   ": "Günlük/aylık görünümü değiştir",
        "   h   ": "Etkinliği yüksek öncelikli olarak değiştir",
        "   l   ": "Etkinliği düşük öncelikli olarak değiştir",
        "   d   ": "Toggle event as done",
        "   .   ": "Etkinlik gizliliğini aç / kapat",
        "   C   ": "Calcurse'den etkinlikleri içe aktarma",
        "   R   ": "Geçerli aya (güne) dön",
        }

KEYS_TODO = {
        "  a(A) ": "Yeni (alt) görev ekle",
        "  h(H) ": "Görevlerden birini (tümünü) yüksek öncelikli olarak değiştirin",
        "  l(L) ": "Görevlerden birini (tümünü) düşük öncelikli olarak değiştirin",
        "  d(D) ": "Görevlerden birini (tümünü) yapıldı olarak değiştirin",
        "  u(U) ": "Görevlerden birinin (tümünün) işaretini kaldırın",
        "  x(X) ": "Görevlerden birini (tümünü) silme (tüm alt görevlerle birlikte)",
        "  t(T) ": "Bir görev için zamanlayıcıyı başlatma/duraklatma (kaldırma)",
        "  e,c  ": "Görev düzenleme",
        "   s   ": "Görev ve alt görev arasında geçiş yapma",
        "   .   ": "Görev gizliliğini aç / kapat",
        "  f(F) ": "Görev son tarihini değiştir (kaldır)",
        "   m   ": "Görev taşıma",
        "  C(W) ": "Calcurse'den görevleri içe aktarma (taskwarrior)",
        }

MSG_NAME          = "CALCURE"
MSG_VIM           = "Vim tuşları (j, k, ZZ, ZQ) da çalışıyor!"
MSG_INFO          = "Daha fazla bilgi için ziyaret edin:"
MSG_SITE          = "https://anufrievroman.gitbook.io/calcure"
MSG_EXIT          = "Gerçekten çıkıyorsun musun? (y/n) "

MSG_EVENT_HIGH    = "Yüksek öncelikli olay numarası olarak işaretle: "
MSG_EVENT_LOW     = "Düşük öncelikli olay numarası olarak işaretle: "
MSG_EVENT_DONE    = "Mark as done event number: "
MSG_EVENT_RESET   = "Etkinlik numarası için durumu sıfırla: "
MSG_EVENT_DEL     = "Etkinlik numarasını sil: "
MSG_EVENT_REN     = "Etkinlik numarasını yeniden adlandır: "
MSG_NEW_TITLE     = "Yeni başlık girin: "
MSG_EVENT_MV      = "Etkinlik numarasını taşı: "
MSG_EVENT_MV_TO   = "Etkinliği şuraya taşı (YYYY/MM/DD): "
MSG_EVENT_MV_TO_D = "Etkinliği şuraya taşı: "
MSG_EVENT_DATE    = "Tarih girin: "
MSG_EVENT_TITLE   = "Başlığı girin: "
MSG_EVENT_REP     = "Etkinliğin kaç kez tekrarlanacağı: "
MSG_EVENT_FR      = "Etkinliği her gün(d), hafta(w), ay(m) veya yıl(y) tekrarlayın? "
MSG_EVENT_IMP     = "Calcurse'den etkinlikleri içe aktarma? (y/n)"
MSG_EVENT_PRIVACY = "Etkinlik numarasının gizliliğini aç / kapat: "
MSG_TM_ADD        = "Görev numarası için zamanlayıcı ekle/duraklat: "
MSG_TM_RESET      = "Görev numarası için zamanlayıcıyı kaldır: "
MSG_TS_HIGH       = "Görev numarasını yüksek öncelikli olarak işaretleyin: "
MSG_TS_LOW        = "Görev numarasını düşük öncelikli olarak işaretleyin: "
MSG_TS_RES        = "Görev numarasının durumunu sıfırla: "
MSG_TS_DONE       = "Görev numarasını tamamlandı olarak işaretle: "
MSG_TS_DEL        = "Görev numarasını sil: "
MSG_TS_DEL_ALL    = "Gerçekten tüm görevleri silecek misiniz? (y/n)"
MSG_TS_MOVE       = "Numaradan görevi taşı: "
MSG_TS_MOVE_TO    = "Görevi numaraya taşı: "
MSG_TS_EDIT       = "Görev numarasını düzenle: "
MSG_TS_TOG        = "Alt görev numarasını değiştir: "
MSG_TS_SUB        = "Görev numarası için alt görev ekleyin: "
MSG_TS_TITLE      = "Alt görev girin: "
MSG_TS_IM         = "Calcurse'den görevleri içe aktarma? (y/n)"
MSG_TS_TW         = "Taskwarrior'dan görevleri içe aktarma? (y/n)"
MSG_TS_NOTHING    = "Planlanmış bir şey yok..."
MSG_TS_PRIVACY    = "Görev numarasının gizliliğini aç / kapat: "
MSG_TS_DEAD_ADD   = "Görev numarası için son tarih ekle: "
MSG_TS_DEAD_DEL   = "Görev numarasının son tarihini kaldırın: "
MSG_TS_DEAD_DATE  = "Son tarih ekleyin (YYYY/MM/DD): "
MSG_WEATHER       = "Hava durumu yükleniyor..."
MSG_ERRORS        = "Errors have occurred. See info.log in your config folder."
MSG_GOTO          = "Go to date (YYYY/MM/DD): "
MSG_GOTO_G        = "Go to date: "
MSG_INPUT         = "Incorrect input."

CALENDAR_HINT     = "Space · Günlüğe geç   a · Etkinlik ekle  n/p · Ayı değiştir   ? · Tüm tuş atamaları"
CALENDAR_HINT_D   = "Space · Günlüğe geç   a · Etkinlik ekle  n/p · Günü değiştir   ? · Tüm tuş atamaları"
JOURNAL_HINT      = "Space · Takvime geç   a · Görev ekle   d · Tamamlandı   i · Önemli   ? · Tüm tuş atamaları"

DAYS = ["PAZARTESİ", "SALI", "ÇARŞAMBA", "PERŞEMBE", "CUMA", "CUMARTESİ", "PAZAR"]
DAYS_PERSIAN = ["SHANBEH", "YEKSHANBEH", "DOSHANBEH", "SESHANBEH", "CHAHARSHANBEH", "PANJSHANBEH", "JOMEH"]

MONTHS = ["OCAK", "ŞUBAT", "MART", "NİSAN", "MAYIS", "HAZİRAN", "TEMMUZ", "AĞUSTOS", "EYLÜL", "EKİM", "KASIM", "ARALIK"]
MONTHS_PERSIAN = ["FARVARDIN", "ORDIBEHESHT", "KHORDAD", "TIR", "MORDAD", "SHAHRIVAR", "MEHR", "ABAN", "AZAR", "DEY", "BAHMAN", "ESFAND"]



"""Module that controls saving data files"""

import os

from calcure.data import *
from calcure.calendars import convert_to_persian_date, convert_to_gregorian_date


class TaskSaverCSV:
    """Save tasks into CSV files"""

    def __init__(self, user_tasks, cf):
        self.user_tasks = user_tasks
        self.tasks_file = cf.TASKS_FILE
        self.use_persian_calendar = cf.USE_PERSIAN_CALENDAR

    def save(self):
        """Rewrite CSV file with changed tasks"""
        original_file = self.tasks_file
        dummy_file = self.tasks_file + '.bak'
        with open(dummy_file, "w", encoding="utf-8") as f:
            for task in self.user_tasks.items:

                # If persian calendar was used, we convert event back to Gregorian for storage:
                if self.use_persian_calendar and task.year != 0:
                    year, month, day = convert_to_gregorian_date(task.year, task.month, task.day)
                else:
                    year, month, day = task.year, task.month, task.day

                dot = "."
                f.write(f'{year},{month},{day},"{dot*task.privacy}{task.name}",{task.status.name.lower()}')
                for stamp in task.timer.stamps:
                    f.write(f',{str(stamp)}')
                f.write("\n")
        os.remove(original_file)
        os.rename(dummy_file, original_file)
        self.user_tasks.changed = False


class EventSaverCSV:
    """Save events into CSV files"""

    def __init__(self, user_events, cf):
        self.user_events = user_events
        self.events_file = cf.EVENTS_FILE
        self.use_persian_calendar = cf.USE_PERSIAN_CALENDAR

    def save(self):
        """Rewrite the data file with changed events"""
        original_file = self.events_file
        dummy_file = self.events_file + '.bak'
        with open(dummy_file, "w", encoding="utf-8") as file:
            for ev in self.user_events.items:

                # If persian calendar was used, we convert event back to Gregorian for storage:
                if self.use_persian_calendar:
                    year, month, day = convert_to_gregorian_date(ev.year, ev.month, ev.day)
                else:
                    year, month, day = ev.year, ev. month, ev.day

                name = f'{"."*ev.privacy}{ev.name}'
                file.write(f'{ev.item_id},{year},{month},{day},"{name}",{ev.repetition},{ev.frequency.name.lower()},{ev.status.name.lower()}\n')
        os.remove(original_file)
        os.rename(dummy_file, original_file)
        self.user_events.changed = False
